{ ------------------------------------------------------------------------------------------------------------------- }
{                                                                                                                     }
{ Die Unit stellt einen inhaltslosen Datentypen TAVLBaum zur Verfgung. Um einen AVL-Baum mit Inhalt zu erhalten,     }
{ muss eine hiervon abgeleitete Klasse erstellt werden, z. B.:                                                        }
{                                                                                                                     }
{  type TIntAVLBaum = class(TAVLBaum)                                                                                 }
{         public                                                                                                      }
{           Wurzelinhalt: integer;                                                                                    }
{           constructor Create(w: Integer);                                                                           }
{           function equal(e: TSuchbaum): boolean; override;        // Diese stammt noch von TSuchbaum                }
{           function higher(e: TSuchbaum): boolean; override;       // Diese ebenfalls                                }
{           function WurzelinhaltToString: string; override;                                                          }
{       end;                                                                                                          }
{                                                                                                                     }
{  Die Routinen equal (gleich) und higher (grer) dienen zur Einsortierung in den Such- bzw. AVL-Baum.               }
{                                                                                                                     }
{ -------------------------------------------------------------------------- Tool zum AVL-Baum, (C) 2002 D. Garmann - }

unit Avlbaum;

interface

uses ExtCtrls, suchbaum;

type TAVLBaum=class(TSuchBaum)
       private
         Hoehe: integer;
           { Damit die Hhe nicht immer rekursiv berechnet werden muss, wird sie lokal in jedem Knoten gespeichert    }

         function IsoliereGroesstesElement(var grElem: TAVLbaum): TAVLbaum;
           { Liefert den AVLBaum ohne das groesste Element. Dieses steht abgekapselt in grElement                     }
         function EchtLoeschen: TAVLbaum;
           { Liefert den neuen AVLbaum ohne das zuvor existierende (!!!) Suchelement e                                }
         function AVLKorrektur: TAVLBaum;
           { Liefert den - falls ntig - hhenbalancierten AVLbaum                                                    }
         procedure Aktualisiere_Hoehe;
           { Aktualisiert die Variable Hoehe                                                                          }

       public
         constructor create;
         function Balance: integer;
           { Liefert die Balance (rechterTeilbaum.Hoehe - linkerTeilbaum.Hoehe)                                       }
         function BaumHoehe: integer;
           { Liefert den Wert der privaten Variablen Hoehe                                                            }
         function AVLEinfuegen(e: TAVLBaum; var ExistiertBereits: boolean): TAVLBaum;
           { Liefert den neuen AVLBaum mit dem eingefgten Suchelement                                                }
         function AVLLoeschen(e: TAVLBaum; var Existiert: boolean): TAVLBaum;
           { Liefert den neuen AVLBaum ohne das eventuell vorhandene Suchelement e                                    }
           { war es nicht vorhanden => Existiert = false                                                              }

     end;

implementation

constructor TAVLBaum.create;
begin
  inherited create;
  Hoehe:= 1;
end;

function TAVLBaum.Balance: integer;
begin
  if leer
    then Balance:= 0
    else Balance:= (rechterTeilbaum as TAVLBaum).Baumhoehe - (linkerTeilbaum as TAVLBaum).Baumhoehe;
end;

function TAVLBaum.Baumhoehe: integer;
begin
  if leer
    then Baumhoehe:= 0
    else Baumhoehe:= Hoehe;
end;

procedure TAVLBaum.Aktualisiere_Hoehe;
begin
  if not leer
  then
  if (rechterTeilbaum as TAVLBaum).Baumhoehe >
     (linkerTeilbaum as TAVLBaum).Baumhoehe
       then Hoehe:= (rechterTeilbaum as TAVLBaum).Baumhoehe + 1
       else Hoehe:= (linkerTeilbaum as TAVLBaum).Baumhoehe + 1;
end;

function TAVLBaum.AVLKorrektur: TAVLBaum;
var x,y,z, b1, b2, b3, b4: TAVLBaum;
begin
  case Balance of
   -1,0,1: AVLKorrektur:= Self;                                          { Bei Balance in -1,0,1 ist alles ok.        }
       -2: begin                                                         { Linkslastig:                               }
             case (LinkerTeilbaum as TAVLBaum).Balance of
               -1: begin                                                 {   stark Auenlastig:                       }
                     x:= Self;                                           {     dann RR mit Balance und Hhenausgleich }
                     y:= x.linkerTeilbaum as TAVLBaum;
                     b2:= y.rechterTeilbaum as TAVLBaum;
                     x.FuegeTeilbaumLinksAn(b2);
                     y.FuegeTeilbaumRechtsAn(x);
                     x.Hoehe:= x.Hoehe-2;
                     y.Hoehe:= y.Hoehe+ 0;
                     AVLKorrektur:= y;
                   end;
                0: begin                                                 {   schwach Auenlastig:                     }
                     x:= Self;                                           {     dann RR mit Balance und Hhenausgleich }
                     y:= x.linkerTeilbaum as TAVLBaum;
                     b2:= y.rechterTeilbaum as TAVLBaum;
                     x.FuegeTeilbaumLinksAn(b2);
                     y.FuegeTeilbaumRechtsAn(x);
                     x.Hoehe:= x.Hoehe-1;
                     y.Hoehe:= y.Hoehe+ 1;
                     AVLKorrektur:= y;
                   end;
               +1: begin                                                 {   Innenlastig:                             }
                     x:= Self;                                           {     dann LR, und dann RR mit               }
                     y:= x.linkerTeilbaum as TAVLBaum;                   {     Balance und Hhenausgleich             }
                     z:= y.rechterTeilbaum as TAVLBaum;
                     b2:= z.linkerTeilbaum as TAVLBaum;
                     b3:= z.rechterTeilbaum as TAVLBaum;
                     y.FuegeTeilbaumRechtsAn(b2);
                     z.FuegeTeilbaumLinksAn(y);
                     x.FuegeTeilbaumLinksAn(b3);
                     z.FuegeTeilbaumRechtsAn(x);
                     x.Hoehe:= x.Hoehe- 2;
                     y.Hoehe:= y.Hoehe- 1;
                     z.Hoehe:= z.Hoehe+ 1;
                     AVLKorrektur:= z;
                   end;
             end;
           end;
       +2: begin
             case (RechterTeilbaum as TAVLBaum).Balance of
               +1: begin
                     x:= Self;
                     y:= x.rechterTeilbaum as TAVLBaum;
                     b2:= y.linkerTeilbaum as TAVLBaum;
                     x.FuegeTeilbaumRechtsAn(b2);
                     y.FuegeTeilbaumLinksAn(x);
                     x.Hoehe:= x.Hoehe-2;
                     y.Hoehe:= y.Hoehe+ 0;
                     AVLKorrektur:= y;
                   end;
                0: begin
                     x:= Self;
                     y:= x.rechterTeilbaum as TAVLBaum;
                     b2:= y.linkerTeilbaum as TAVLBaum;
                     x.FuegeTeilbaumRechtsAn(b2);
                     y.FuegeTeilbaumLinksAn(x);
                     x.Hoehe:= x.Hoehe-1;
                     y.Hoehe:= y.Hoehe+ 1;
                     AVLKorrektur:= y;
                   end;
               -1: begin
                     x:= Self;
                     y:= x.rechterTeilbaum as TAVLBaum;
                     z:= y.linkerTeilbaum as TAVLBaum;
                     b2:= z.linkerTeilbaum as TAVLBaum;
                     b3:= z.rechterTeilbaum as TAVLBaum;
                     y.FuegeTeilbaumLinksAn(b3);
                     z.FuegeTeilbaumRechtsAn(y);
                     x.FuegeTeilbaumRechtsAn(b2);
                     z.FuegeTeilbaumLinksAn(x);
                     x.Hoehe:= x.Hoehe- 2;
                     y.Hoehe:= y.Hoehe- 1;
                     z.Hoehe:= z.Hoehe+ 1;
                     AVLKorrektur:= z;
                   end;
             end;
           end;
  end;
end;

function TAVLBaum.AVLEinfuegen(e: TAVLBaum; var ExistiertBereits: boolean): TAVLBaum;
var tb: TAVLBaum;
    hoehe_l, hoehe_r: integer;
begin
  if Leer                                                                { Baum ist leer, dann                        }
  then begin
         AVLEinfuegen:= e;                                               { wird das Element e zum neuen Suchbaum      }
         ExistiertBereits:= false;                                       { und war natrlich nicht vorhanden          }
       end
  else begin
         If equal(e)                                                     { Wenn die aktuelle Wurzel gleich e ist,     }
         then begin
                ExistiertBereits:= true;                                 { dann kann es nicht eingefgt werden        }
                AVLeinfuegen:= Self;                                     { Gib also unvernderten Baum zurck         }
              end
         else begin
                if higher(e)                                             { Wenn aktuelle Wurzel grter als e ist,    }
                then begin
                       tb:=(LinkerTeilbaum as TAVLbaum);                 { dann muss das Element in den linken Teil-  }
                       tb:= tb.AVLeinfuegen(e, ExistiertBereits);        { baum eingefgt werden. Dieser (evtl.) er-  }
                       FuegeTeilbaumLinksAn(tb);                         { weiterte linke Teilbaum wird links angehngt}
                     end
                else begin
                       tb:= (RechterTeilbaum as TAVLbaum);               { das Element muss in den rechten Teilbaum   }
                       tb:= tb.AVLeinfuegen(e, ExistiertBereits);        { eingefgt werden. Dieser (evtl.) erweiterte}
                       FuegeTeilbaumRechtsAn(tb);                        { rechte Teilbaum wird links angehngt und   }
                     end;
                Aktualisiere_Hoehe;                                      { Wenn eingefgt wurde, so hat sich die Hhe }
                                                                         { eventuell gendert. Also Neu berechnen     }
                AVLeinfuegen:= AVLKorrektur;                             { Baum eventuell ausgleichen. Beim Ausgleich }
                                                                         { kann es passieren, dass ein anderes Elem.  }
                                                                         { zur neuen Wurzel wird. Deshalb wird das    }
                                                                         { Funktionsergebnis weiter nach oben gereicht}
              end;
       end;
end;


    function TAVLBaum.IsoliereGroesstesElement(var grElem: TAVLBaum): TAVLBaum;
    var tb: TAVLBaum;
    begin
      if RechtsLeer
      then begin                                                         { grtes Element gefunden, also:            }
             IsoliereGroesstesElement:= LinkerTeilbaum as TAVLBaum;      { dieses Element isolieren, d. h. den linken }
                                                                         { Teilbaum zurckliefern.                    }
             grElem:= self;                                              { und das grte Element in CbR-Variablen    }
                                                                         { zurckgeben                                }
           end
      else begin                                                         { noch kein grtes Element                  }
             tb:= RechterTeilbaum as TAVLBaum;                           { also muss die Suche im rechten Teilbaum    }
             tb:= tb.IsoliereGroesstesElement(grElem);                   { weitergehen. Dieser wird anschlieend ohne }
             FuegeTeilbaumRechtsAn(tb);                                  { das grte Element wieder angehngt.       }
             tb:= AVLKorrektur;                                          { tb auf den Hhenbalancierten Baum          }
             tb.Aktualisiere_Hoehe;                                      {    und noch die Hhe aktualisieren         }
             IsoliereGroesstesElement:= tb;                              { und der vernderte Baum wird zurckgegeben }
           end;
    end;

  function TAVLBaum.EchtLoeschen: TAVLBaum;
  var grElement: TAVLBaum;                                               { grtes Element im linken Teilbaum         }
      tb: TAVLBaum;
  begin
    if RechtsLeer                                                        { wenn rechter Teilbaum leer ist,            }
    then begin
           tb:= ErsetztDurchLinkenTeilbaum as TAVLBaum;                  {    dann durch diesen ersetzten.            }
           EchtLoeschen:= tb.AVLKorrektur;                               {    und eventuell ausgleichen               }
         end
    else if LinksLeer                                                    { wenn linker Teilbaum leer ist,             }
         then begin
                tb:= ErsetztDurchRechtenTeilbaum as TAVLBaum;            {    dann durch diesen ersetzten.            }
                EchtLoeschen:= tb.AVLKorrektur;                          {    und eventuell ausgleichen               }
              end
         else begin
                tb:= LinkerTeilbaum as TAVLBaum;                         { Linken Teilbaum holen                      }
                tb:= tb.IsoliereGroesstesElement(grElement);             { in diesem das grte Element isolieren     }
                FuegeTeilbaumLinksAn(tb);                                { und diesen Teilbaum wieder anhngen        }

                grElement.FuegeTeilbaumLinksAn(LinkerTeilbaum);          { hnge jetzt aktuellen linken und rechten TB}
                grElement.FuegeTeilbaumRechtsAn(RechterTeilbaum);        { an das zuvor ermittelte grte Element an  }
                grElement.Aktualisiere_Hoehe;
                FuegeTeilbaumLinksAn(NIL);                               { und trenne die Teilbume vom aktuellen     }
                FuegeTeilbaumRechtsAn(NIL);                              { Element.                                   }
                Free;                                                    { Jetzt kann das aktuelle Element gelscht   }
                EchtLoeschen:= grElement;                                { werden und das zuvor grte Element zurck }
              end;                                                       { gegeben werden.                            }
  end;


function TAVLBaum.AVLLoeschen(e: TAVLBaum; var Existiert: boolean): TAVLBaum;
var tb: TAVLBaum;
    hoehe_l, hoehe_r: integer;
begin
  If Leer
  then begin                                                             { Baum ist leer, dann                        }
         Existiert:= false;                                              { kann man nichts lschen.                   }
         AVLLoeschen:= Self;                                             { Gib den unvernderten Baum zurck          }
       end
  else begin
         if equal(e)                                                     { Wenn das Suchelement existiert, dann       }
         then begin
                tb:= EchtLoeschen as TAVLBaum;                           { aktuellen Baum ohne das Element in Variable}
                Existiert:= true;                                        { tb zwischenspeichern.                      }
                if not tb.leer                                           { In der Wurzel kann die Balance gestrt     }
                  then tb:= tb.AVLKorrektur;                             { sein. Diese eventuell ausgleichen          }
                AVLLoeschen:= tb;                                        { Ist alles erledigt, dann ist tb der akt.   }
                                                                         { Baum und kann zurckgegeben werden.        }
              end
         else begin
                if higher(e)                                             { Wenn die aktuelle Wurzel grer als e, dann}
                then begin
                       tb:= LinkerTeilbaum as TAVLbaum;                  { Suche im linken Teilbaum weiter und lsche }
                       tb:= tb.AVLLoeschen(e, Existiert);                { darin das Element.                         }
                       FuegeTeilbaumLinksAn(tb);                         { Hnge anschlieend den vernderten Baum    }
                                                                         { links an.                                  }
                       Aktualisiere_Hoehe;                               { Wenn gelscht wurde, so hat sich die Hhe  }
                                                                         { eventuell gendert. Also neu berechnen     }
                     end
                else begin
                       tb:= RechterTeilbaum as TAVLbaum;                 { Suche im rechten Teilbaum weiter und lsche}
                       tb:= tb.AVLLoeschen(e, Existiert);                { darin das Element.                         }
                       FuegeTeilbaumRechtsAn(tb);                        { Hnge anschlieend den vernderten Baum    }
                                                                         { rechts an.                                 }
                       Aktualisiere_Hoehe;                               { Wenn gelscht wurde, so hat sich die Hhe  }
                                                                         { eventuell gendert. Also neu berechnen     }
                     end;
                AVLLoeschen:= AVLKorrektur;                              { Baum eventuell ausgleichen. Beim Ausgleich }
                                                                         { kann es passieren, dass ein anderes Elem.  }
                                                                         { zur neuen Wurzel wird. Deshalb wird das    }
                                                                         { Funktionsergebnis weiter nach oben gereicht}
              end;
       end;
end;

end.
