{ ------------------------------------------------------------------------------------------------------------------- }
{                                                                                                                     }
{ Die Unit stellt einen inhaltslosen Datentypen TSuchBaum zur Verfgung. Um einen Suchbaum mit Inhalt zu erhalten,    }
{ muss eine hiervon abgeleitete Klasse erstellt werden, z. B.:                                                        }
{                                                                                                                     }
{  type TIntSuchbaum = class(TSuchbaum)                                                                               }
{         public                                                                                                      }
{           Wurzelinhalt: integer;                                                                                    }
{           constructor Create(w: Integer);                                                                           }
{           function equal(e: TSuchbaum): boolean; override;                                                          }
{           function higher(e: TSuchbaum): boolean; override;                                                         }
{           function WurzelinhaltToString: string; override;                                                          }
{       end;                                                                                                          }
{                                                                                                                     }
{  Die Routinen equal (gleich) und higher (grer) dienen zur Einsortierung in den Suchbaum.                          }
{                                                                                                                     }
{ -------------------------------------------------------------------------- Tool zum Suchbaum, (C) 2001 D. Garmann - }

unit Suchbaum;

interface

uses ExtCtrls, binbaum;

type TSuchBaum=class(TBinBaum)
       private
       function IsoliereGroesstesElement(var grElem: TSuchbaum): TSuchbaum;
         { Liefert den SuchBaum ohne das groesste Element. Dieses steht abgekapselt in grElement                      }
       function EchtLoeschen: TSuchbaum;
         { Liefert den neuen Suchbaum ohne das zuvor existierende (!!!) Suchelement e                                 }

       public
         { mssen fr einen neuen Suchbaum-Typen definiert werden                                                     }
       function equal(e: TSuchbaum): boolean; virtual; abstract;
       function higher(e: TSuchbaum): boolean; virtual; abstract;
         { werden aus den beiden oberen Funktionen abgeleitet                                                         }
       function notequal(e: TSuchbaum): boolean;
       function higherequal(e: TSuchbaum): boolean;
       function lower(e: TSuchbaum): boolean;
       function lowerequal(e: TSuchbaum): boolean;

       function SuchEinfuegen(e: TSuchbaum; var ExistiertBereits: boolean): TSuchbaum;
         { Liefert den neuen SuchBaum mit dem eingefgten Suchelement                                                 }
       function SuchLoeschen(e: TSuchbaum; var Existiert: boolean): TSuchbaum;
         { Liefert den neuen Suchbaum ohne das eventuell vorhandene Suchelement e                                     }
         { war es nicht vorhanden => Existiert = false                                                                }
     end;

implementation

function TSuchBaum.notequal(e: TSuchbaum): boolean;
begin
  notequal:= not equal(e);
end;

function TSuchBaum.higherequal(e: TSuchbaum): boolean;
begin
  higherequal:= higher(e) or equal(e);
end;

function TSuchBaum.lower(e: TSuchbaum): boolean;
begin
  lower:= not higherequal(e);
end;

function TSuchBaum.lowerequal(e: TSuchbaum): boolean;
begin
  lowerequal:= lower(e) or equal(e);
end;

function TSuchbaum.SuchEinfuegen(e: TSuchbaum; var ExistiertBereits: boolean): TSuchbaum;
var tb: TSuchBaum;
begin
  if Leer                                                                { Baum ist leer, dann                        }
  then begin
         SuchEinfuegen:= e;                                              { wird das Element e zum neuen Suchbaum      }
         ExistiertBereits:= false;                                       { und war natrlich nicht vorhanden          }
       end
  else begin
         If equal(e)                                                     { Wenn die aktuelle Wurzel gleich e ist,     }
         then begin
                ExistiertBereits:= true;                                 { dann kann es nicht eingefgt werden        }
                Sucheinfuegen:= Self;                                    { Gib also unvernderten Baum zurck         }
              end
         else if higher(e)                                               { Wenn aktuelle Wurzel grter als e ist,    }
              then begin
                     tb:=(LinkerTeilbaum as TSuchbaum);                  { dann muss das Element in den linken Teil-  }
                     tb:= tb.Sucheinfuegen(e, ExistiertBereits);         { baum eingefgt werden. Dieser (evtl.) er-  }
                     FuegeTeilbaumLinksAn(tb);                           { weiterte linke Teilbaum wird links ange-   }
                     Sucheinfuegen:= self;                               { hngt und der neue Baum wird zurckgegeben }
                   end
              else begin
                     tb:= (RechterTeilbaum as TSuchbaum);                { das Element muss in den rechten Teilbaum   }
                     tb:= tb.Sucheinfuegen(e, ExistiertBereits);         { eingefgt werden. Dieser (evtl.) erweiterte}
                     FuegeTeilbaumRechtsAn(tb);                          { rechte Teilbaum wird links angehngt und   }
                     Sucheinfuegen:= self;                               { der so erweiterte Baum wird zurckgegeben. }
                   end
       end;
end;

    function TSuchbaum.IsoliereGroesstesElement(var grElem: TSuchbaum): TSuchbaum;
    var tb: TSuchbaum;
    begin
      if RechtsLeer
      then begin                                                         { grtes Element gefunden, also:            }
             IsoliereGroesstesElement:= LinkerTeilbaum as TSuchbaum;     { dieses Element isolieren, d. h. den linken }
                                                                         { Teilbaum zurckliefern.                    }
             grElem:= self;                                              { und das grte Element in CbR-Variablen    }
                                                                         { zurckgeben                                }
           end
      else begin                                                         { noch kein grtes Element                  }
             tb:= RechterTeilbaum as TSuchbaum;                          { also muss die Suche im rechten Teilbaum    }
             tb:= tb.IsoliereGroesstesElement(grElem);                   { weitergehen. Dieser wird anschlieend ohne }
             FuegeTeilbaumRechtsAn(tb);                                  { das grte Element wieder angehngt.       }
             IsoliereGroesstesElement:= self;                            { und der vernderte Baum wird zurckgegeben }
           end;
    end;

  function TSuchbaum.EchtLoeschen: TSuchbaum;
  var grElement: TSuchbaum;                                              { grtes Element im linken Teilbaum         }
      tb: TSuchbaum;
  begin
    if RechtsLeer                                                        { wenn rechter Teilbaum leer ist,            }
    then EchtLoeschen:= ErsetztDurchLinkenTeilbaum as TSuchbaum          {    dann durch diesen ersetzten.            }
    else if LinksLeer                                                    { wenn linker Teilbaum leer ist,             }
         then EchtLoeschen:= ErsetztDurchRechtenTeilbaum as TSuchbaum    {    dann durch diesen ersetzten.            }
         else begin
                tb:= LinkerTeilbaum as TSuchbaum;                        { Linken Teilbaum holen                      }
                tb:= tb.IsoliereGroesstesElement(grElement);             { in diesem das grte Element isolieren     }
                FuegeTeilbaumLinksAn(tb);                                { und diesen Teilbaum wieder anhngen        }

                grElement.FuegeTeilbaumLinksAn(LinkerTeilbaum);          { hnge jetzt aktuellen linken und rechten TB}
                grElement.FuegeTeilbaumRechtsAn(RechterTeilbaum);        { an das zuvor ermittelte grte Element an  }
                FuegeTeilbaumLinksAn(NIL);                               { und trenne die Teilbume vom aktuellen     }
                FuegeTeilbaumRechtsAn(NIL);                              { Element.                                   }
                Free;                                                    { Jetzt kann das aktuelle Element gelscht   }
                EchtLoeschen:= grElement;                                { werden und das zuvor grte Element zurck }
              end;                                                       { gegeben werden.                            }
  end;


function TSuchbaum.SuchLoeschen(e: TSuchbaum; var Existiert: boolean): TSuchbaum;
var tb: TSuchBaum;
begin
  If Leer
  then begin                                                             { Baum ist leer, dann                        }
         Existiert:= false;                                              { kann man nichts lschen.                   }
         SuchLoeschen:= Self;                                            { Gib den unvernderten Baum zurck          }
       end
  else begin
         if equal(e)                                                     { Wenn das Suchelement existiert, dann       }
         then begin
                SuchLoeschen:= EchtLoeschen;                             { aktuellen Baum ohne das Element e zurck-  }
                Existiert:= true;                                        { geben.                                     }
              end
         else if higher(e)                                               { Wenn die aktuelle Wurzel grer als e, dann}
              then begin
                     tb:= LinkerTeilbaum as TSuchbaum;                   { Suche im linken Teilbaum weiter und lsche }
                     tb:= tb.SuchLoeschen(e, Existiert);                 { darin das Element.                         }
                     FuegeTeilbaumLinksAn(tb);                           { Hnge anschlieend den vernderten Baum    }
                     SuchLoeschen:= Self;                                { links an und gib den gesamten Baum zurck. }
                   end
              else begin
                     tb:= RechterTeilbaum as TSuchbaum;                  { Suche im rechten Teilbaum weiter und lsche}
                     tb:= tb.SuchLoeschen(e, Existiert);                 { darin das Element.                         }
                     FuegeTeilbaumRechtsAn(tb);                          { Hnge anschlieend den vernderten Baum    }
                     SuchLoeschen:= Self;                                { rechts an und gib den gesamten Baum zurck.}
                   end;
       end;
end;

end.
